/*  Copyright (C) 2007 University of Oxford  */

/* S. Jbabdi */

/*  Part of FSL - FMRIB's Software Library
    http://www.fmrib.ox.ac.uk/fsl
    fsl@fmrib.ox.ac.uk

    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK


    LICENCE

    FMRIB Software Library, Release 6.0 (c) 2018, The University of
    Oxford (the "Software")

    The Software remains the property of the Oxford University Innovation
    ("the University").

    The Software is distributed "AS IS" under this Licence solely for
    non-commercial use in the hope that it will be useful, but in order
    that the University as a charitable foundation protects its assets for
    the benefit of its educational and research purposes, the University
    makes clear that no condition is made or to be implied, nor is any
    warranty given or to be implied, as to the accuracy of the Software,
    or that it will be suitable for any particular purpose or for use
    under any specific conditions. Furthermore, the University disclaims
    all responsibility for the use which is made of the Software. It
    further disclaims any liability for the outcomes arising from using
    the Software.

    The Licensee agrees to indemnify the University and hold the
    University harmless from and against any and all claims, damages and
    liabilities asserted by third parties (including claims for
    negligence) which arise directly or indirectly from the use of the
    Software or the sale of any products based on the Software.

    No part of the Software may be reproduced, modified, transmitted or
    transferred in any form or by any means, electronic or mechanical,
    without the express permission of the University. The permission of
    the University is not required if the said reproduction, modification,
    transmission or transference is done without financial return, the
    conditions of this Licence are imposed upon the receiver of the
    product, and all original and amended source code is included in any
    transmitted product. You may be held legally responsible for any
    copyright infringement that is caused or encouraged by your failure to
    abide by these terms and conditions.

    You are not permitted under this Licence to use this Software
    commercially. Use for which any financial return is received shall be
    defined as commercial use, and includes (1) integration of all or part
    of the source code or the Software into a product for sale or license
    by or on behalf of Licensee to third parties or (2) use of the
    Software or any derivative of it for research with the final aim of
    developing software products for sale or license to a third party or
    (3) use of the Software or any derivative of it for research with the
    final aim of developing non-software products for sale or license to a
    third party, or (4) use of the Software to provide any service to an
    external organisation for which payment is received. If you are
    interested in using the Software commercially, please contact Oxford
    University Innovation ("OUI"), the technology transfer company of the
    University, to negotiate a licence. Contact details are:
    fsl@innovation.ox.ac.uk quoting Reference Project 9564, FSL.*/

#include <cmath>
#include <stdlib.h>

#include "utils/options.h"
#include "newimage/newimageall.h"
#include "miscmaths/miscmaths.h"
#include "miscmaths/miscprob.h"


using namespace Utilities;
using namespace NEWIMAGE;
using namespace NEWMAT;
using namespace MISCMATHS;

string title="dpm_preprocess\n create data file for dpm clustering";
string examples="dpm_preprocess -s <seed> -t <targetfile> -o basename [--rand --reduce=10.0]";

Option<string> seed(string("-s,--seed"),"",
		       string("Seed mask volume"),
		       true,requires_argument);
Option<string> targets(string("-t,--targets"),"",
		       string("Targets file as in seed_to_target"),
		       true,requires_argument);
Option<string> directory(string("-d,--dir"),"",
		       string("probtrackx results directory"),
		       true,requires_argument);
Option<string> output(string("-o,--out"),"",
		       string("Ouput file name"),
		       true,requires_argument);
Option<bool>   addrand(string("--rand"),false,
		       string("Add random noise"),
		       false,no_argument);
Option<int>  dimreduce(string("--reduce"),0,
		       string("Number of dimensions to keep"),
		       false,requires_argument);


void read_masks(vector<string>& masks,const string& filename){
  ifstream fs(filename.c_str());
    string tmp;
    if(fs){
      fs>>tmp;
      do{
	masks.push_back(tmp);
	fs>>tmp;
      }while(!fs.eof());
    }
    else{
      cerr<<filename<<" does not exist"<<endl;
      exit(0);
    }
}
// calculate pca decomposition using the covariance method
// X is the data nxd matrix with n observations (data points) and d variables (dimensions) 
ReturnMatrix pcacov(Matrix& X,const int& dims,Matrix& projector){
  int n=X.Nrows();
  int d=X.Ncols();
  // de-mean
  cout<<"de-mean"<<endl;
  ColumnVector Xmean(d);
  for(int j=1;j<=d;j++)
    Xmean(j) = X.Column(j).Sum()/n;
  for(int j=1;j<=d;j++){
    for(int i=1;i<=n;i++)
      X(i,j) -= Xmean(j);
  }

  // calculate covariance
  cout<<"covariance"<<endl;
  SymmetricMatrix C(d);

  if(d<n)
    C << (X.t()*X)/n;
  else
    C << X*X.t()/n;

  // eigenvalues
  cout<<"eigenvalues"<<endl;
  Matrix V;
  DiagonalMatrix D;
  EigenValues(C,D,V);

  // select subset
  cout<<"variance explained:";
   float cumsum=0,total=D.Trace();
   int dim=0;
   for(int i=D.Nrows();i>=1;i--){
     cumsum += D(i);
     if(dim<=dims){dim++;}
     else{break;}
   }
   cout<<cumsum/total*100<<"%"<<endl;
//   if(dim<=2)dim=2;

//   dim=dims;

  Matrix v(V.Nrows(),dims);
  ColumnVector lam(dims);
  for(int j=1;j<=dims;j++){
    v.Column(j) = V.Column(V.Ncols()-j+1);
    lam(j)=D(V.Ncols()-j+1);
  }

  //cout<<"zscores"<<endl;
  // convert to z-scores
  //for(int i=1;i<=d;i++)
  //X.Row(i) /= sqrt(C(i,i));
  // reconstruct data
  cout<<"data"<<endl;
  Matrix data(dims,n);

  if(!(d<n)){
    v = (X.t() * v);
    for(int j=1;j<=dims;j++)
      v.Column(j) /= sqrt(n*lam(j));
  }
  data = X*v;
  
  projector.ReSize(v.Nrows(),v.Ncols());
  projector = v;

  data.Release();
  return data;
}
ReturnMatrix dimred(Matrix& data,const int& numdims){

  //Matrix projector;
  Matrix newdat,projector;
  newdat=pcacov(data,numdims,projector);
  data.ReSize(newdat.Nrows(),newdat.Ncols());
  data=newdat;

  projector.Release();
  return projector;
}

int preprocess(){
  vector<string> targetnames;
  read_masks(targetnames,targets.value());
  vector< volume<float> > targetmasks;
  volume<float> seedmask;
  read_volume(seedmask,seed.value());  

  volume<float> tmpmask;
  for(unsigned int i=0;i<targetnames.size();i++){
    read_volume(tmpmask,targetnames[i]);
    targetmasks.push_back(tmpmask);
  }

  int npoints=0;
  for(int z=0;z<seedmask.zsize();z++)
    for(int y=0;y<seedmask.ysize();y++)
      for(int x=0;x<seedmask.xsize();x++){
	if(seedmask(x,y,z)==0)continue;
	npoints++;
      }
  Matrix data(npoints,targetmasks.size()),coord(npoints,3);

  volume<float> lookup;
  lookup = seedmask;
  
  int curpoint=0;
  for(int z=0;z<seedmask.zsize();z++)
    for(int y=0;y<seedmask.ysize();y++)
      for(int x=0;x<seedmask.xsize();x++){
	if(seedmask(x,y,z)==0)continue;
	curpoint++;

	for(unsigned int i=0;i<targetmasks.size();i++){
	  float addterm=addrand.value()?0.1*normrnd(1,1).AsScalar():0.0;
	  //OUT(addterm);
	  data(curpoint,i+1) = log(targetmasks[i](x,y,z)+1)+addterm;
	}
	coord(curpoint,1) = x;
	coord(curpoint,2) = y;
	coord(curpoint,3) = z;
  
	lookup(x,y,z) = curpoint;
      }
  cout<<"data created"<<endl;
  // reduce data
  OUT(data.Ncols());
  Matrix proj;
  if(dimreduce.value()>0){
    cout << "reduce dimensionality" << endl;
    proj=dimred(data,dimreduce.value());
  }
  else{
    proj=IdentityMatrix(data.Ncols());
  }
  
  OUT(data.Ncols());
  
  // compute adjacency matrix
  Matrix adj(npoints,26);
  curpoint=0;
  for(int z=0;z<seedmask.zsize();z++)
    for(int y=0;y<seedmask.ysize();y++)
      for(int x=0;x<seedmask.xsize();x++){
	if(seedmask(x,y,z)==0)continue;
	curpoint++;
	short int curneighbour=0;
	for(int zz=-1;zz<=1;zz++)
	  for(int yy=-1;yy<=1;yy++)
	    for(int xx=-1;xx<=1;xx++){
	      if(abs(xx)+abs(yy)+abs(zz)==0)continue;
	      curneighbour++;
	      if(seedmask(x+xx,y+yy,z+zz)!=0)
		adj(curpoint,curneighbour) = lookup(x+xx,y+yy,z+zz);
	      else
		adj(curpoint,curneighbour) = -1;
	    }
	      

      }
  
  
  write_ascii_matrix(proj,output.value()+".proj");
  write_ascii_matrix(data,output.value()+".data");
  write_ascii_matrix(coord,output.value()+".coord");
  write_ascii_matrix(adj,output.value()+".adj");

  return 0;
}


int main (int argc, char *argv[]){

  Tracer tr("main");
  OptionParser options(title,examples);
  
  try{
    options.add(seed);
    options.add(targets);
    options.add(directory);
    options.add(output);
    options.add(addrand);
    options.add(dimreduce);

    options.parse_command_line(argc,argv);

    if ( !options.check_compulsory_arguments(true) ){
      options.usage();
      exit(EXIT_FAILURE);
    }
  }
  catch(X_OptionError& e) {
    options.usage();
    cerr << endl << e.what() << endl;
    exit(EXIT_FAILURE);
  } 
  catch(std::exception &e) {
    cerr << e.what() << endl;
  } 

  return preprocess();

}
